require('anal')
Camera = require('camera')
local game = {}
maxSpeed = 550
local function updateWalkSpeed(stopping)
  local x,y = objects.player.body:getLinearVelocity()
  x,y = math.max(math.min(x, maxSpeed),-maxSpeed),math.max(math.min(y, maxSpeed),-maxSpeed)
  objects.player.body:setLinearVelocity(x,y)
  objects.player.angle = math.atan2(x,-y)
  if stopping then
    objects.player.angle = -objects.player.angle
  end
  local mag = math.abs(x) + math.abs(y)
  if mag < 10 then
    playerLegs:reset()
    playerBody:reset()
  else
    playerLegs:setSpeed(math.min(mag/100,1.5))
    playerBody:setSpeed(math.min(mag/100,1.5))
  end
end

local function createFart()
  local newFart = {}
  local vx,vy = objects.player.body:getLinearVelocity()
  local magnitude = math.sqrt(vx * vx + vy * vy)
  local dx, dy = math.sin(objects.player.angle), math.cos(objects.player.angle)
  local ax,ay
  -- love.graphics.setCaption(#powerups)
  if vx == 0 then
    ax = 0
  else
    ax = vx/math.abs(vx)
  end
  if vy == 0 then
    ay = 0
  else
    ay = vy/math.abs(vy)
  end

  local x, y = objects.player.body:getWorldCenter()
  if lastX then
    x = lastX
  end
  newFart.growthRate = objects.player.fartiness
  newFart.body = love.physics.newBody(world, x+(3*ax), y+(5*ay), "dynamic") --place the body in the center of the world and make it dynamic, so it can move around
  newFart.shape = love.physics.newCircleShape(2) --the player's shape has a radius of 20
  newFart.fixture = love.physics.newFixture(newFart.body, newFart.shape, 1) -- Attach fixture to body and give it a density of 1.
  newFart.fixture:setCategory(2)
  newFart.fixture:setMask(2,6)
  newFart.body:setMass(3*newFart.growthRate)
  newFart.particles = love.graphics.newParticleSystem(fartImages[1], 20*newFart.growthRate)
  newFart.particles:setEmissionRate          (30)
  newFart.particles:setLifetime              (1)
  newFart.particles:setParticleLife          (1,3)
  newFart.particles:setPosition              (0, 0)
  newFart.particles:setDirection             (0)
  newFart.particles:setSpread                (600)
  newFart.particles:setSpeed                 (10, 20, 100*objects.player.speed)
  newFart.particles:setGravity               (0)
  newFart.particles:setRadialAcceleration    (1, 50*newFart.growthRate)
  newFart.particles:setTangentialAcceleration(2)
  newFart.particles:setSizes                 (1,2,3,5*newFart.growthRate)
  newFart.particles:setSizeVariation         (1)
  newFart.particles:setRotation              (0.3)
  newFart.particles:setSpin                  (1)
  newFart.particles:setSpinVariation         (10)
  newFart.particles:setColors                (255,255,255,80, 155,155,155,80, 85,85,85,80)
  local vSpeed
  if not gameOver then
    if ay < 0 then
      vSpeed = vy*65
    else
      vSpeed = vy*65
    end
    newFart.body:applyForce(dx*magnitude*100 + math.random(-100,100),ay*math.abs(dy)*math.max(newFart.growthRate*3,1)*magnitude*65)
  else
    newFart.body:setY(20 + 10*newFart.growthRate)
    newFart.body:applyLinearImpulse(150*newFart.growthRate*math.sin(love.timer.getMicroTime()*50),0)
  end
  newFart.ttl = math.random(300,400)
  table.insert(objects.farts, newFart)
end

local function normalizePlayer()
  objects.player.fartiness = 0.3
  objects.player.speed = 40
  objects.player.scale = 1
  objects.player.body:resetMassData()
end

local function applyEffects(dt)
  normalizePlayer()
  local effects = {
    beer = function() 
      objects.player.angle = -objects.player.angle + math.sin(love.timer.getMicroTime()*20)*0.3 -- wavy farts
    end,
  
    hotdog = function() 
      objects.player.fartiness = 2 -- big farts
    end,

    taco = function()
      objects.player.scale = 2.5
      objects.player.body:setMass(15) -- some kind of fart thing out the back 
      objects.player.speed = 15000
    end,

    curry = function()
      objects.player.remainingFarts = objects.player.maxFarts -- infinite farts
    end,

    medicine = function()
      objects.player.remainingFarts = math.min(objects.player.maxFarts/2, objects.player.remainingFarts) -- gas-b-gone
    end
    
  }
  local effect
  for i=1,#activeEffects do
    if activeEffects[i] then
      effect = activeEffects[i]
      effect.ttl = effect.ttl - dt
      effects[effect.kind]() -- apply
      if effect.ttl < 0 then
        table.remove(activeEffects, i)
      end
    end
  end
end

local function updateFarts(dt)
  local indiciesToRemove = {}
  for i=1, #objects.farts do
    if not objects.farts[i] then 
      break
    end
    if objects.farts[i].ttl < 0 then
      objects.farts[i].body:destroy()
      table.remove(objects.farts, i)
    else
      if objects.farts[i].ttl < 80 then
        objects.farts[i].particles:setSprite(fartImages[3])
      elseif objects.farts[i].ttl < 300 then
        objects.farts[i].particles:setSprite(fartImages[2])
      end
      objects.farts[i].particles:setColors(255,255,255, math.min(objects.farts[i].ttl,80))
      objects.farts[i].ttl = objects.farts[i].ttl - (dt*100)
      objects.farts[i].particles:update(dt)
      local r = objects.farts[i].shape:getRadius() + objects.farts[i].growthRate
      objects.farts[i].shape:setRadius(r)
      objects.farts[i].fixture:getShape():setRadius(r)
    end
  end
end

local function reactToFart(a,b,contact)
  local person
  if a:getCategory() == 1 then
    person = a
  elseif b:getCategory() == 2 then
    person = b
  end

  if not person then return end
  local vx, vy = contact:getNormal()

  person:getBody():applyLinearImpulse(-vx,-vy)
  local idx = person:getUserData().idx
  if not disgustedPatrons[idx] then disgustCount = disgustCount + 1 end
  if (disgustedPatrons[idx] or 0) > 0 then return end -- already said something
  
  local i = math.random(1,#disgustSounds)
  local sound = disgustSounds[i]
  disgustedPatrons[idx] = math.random(1,4)/5
  if sound:isStopped() then
    disgustSounds[i]:setPitch(math.random(7,10)/10)
    disgustSounds[i]:play()
    return
  end
end

local function collideWithPatron()
  collisionCount = collisionCount+1
end

local function consumePowerup(a,b,contact)
  local powerup
  if a:getCategory() == 7 then
    powerup = a
  elseif b:getCategory() == 7 then
    powerup = b
  else
    return
  end
  objects.player.remainingFarts = objects.player.maxFarts
  contact:setEnabled(false)
  love.graphics.setCaption("hit a "..powerup:getUserData().kind)
  table.insert(activeEffects, {kind =  powerup:getUserData().kind, tts = math.random(100,175)/1000, ttl = 5})
  powerups[powerup:getUserData().idx] = nil
  powerup:destroy()
end

local function collide(a,b,contact) -- 1: pedestrian, 2: fart, 3: player, 7: powerup, 5: wall, 9: win
  if gameOver then return end
  local collisionType = a:getCategory() + b:getCategory()
  if collisionType == 3 then -- 1 and 2
    reactToFart(a,b,contact)
  elseif collisionType == 4 then -- 3 and 1
    collideWithPatron()
  elseif collisionType == 10 then -- 3 and 7
    consumePowerup(a,b,contact)
  elseif collisionType == 12 then -- 3 and 9
    gameOver = true
    lastX = objects.player.body:getX()
  end
end

local function placePowerup(x,y,kind)
  local kinds = {'hotdog', 'curry', 'taco', 'beer', 'medicine'}
  local newPowerup,extra = {},{}

  
  newPowerup.body = love.physics.newBody(world, x, y, "kinematic") --place the body in the center of the world and make it dynamic, so it can move around
  newPowerup.shape = love.physics.newCircleShape(20) --the player's shape has a radius of 20
  newPowerup.fixture = love.physics.newFixture(newPowerup.body, newPowerup.shape, 0) -- Attach fixture to body and give it a density of 1.
  newPowerup.fixture:setCategory(7)
  newPowerup.fixture:setMask(1,2)
  extra.kind = kinds[kind+1]
  newPowerup.image = powerupImages[extra.kind]
  extra.idx = #powerups + 1
  newPowerup.fixture:setUserData(extra)
  -- if newPowerup.kind then
  powerups[extra.idx] = newPowerup
  -- end
end

local function createPatron(x,y,image,attitude)
  local patron = {}
  local extra = {}
  extra.animation = newAnimation(image, 64, 64, 0.1, 6)
  extra.angle = math.random(0, 7)
  extra.linePosition = {x,y}
  extra.color = {math.random(200,220),math.random(200,220),math.random(200,220)}
  extra.scale = math.random(4,7)/10
  extra.idx = love.timer.getMicroTime()
  patron.body = love.physics.newBody(world, math.max(x,20), y, "dynamic") --place the body in the center of the world and make it dynamic, so it can move around
  
  if attitude > 0 then
    patron.body:setSleepingAllowed(false)
    patron.body:applyLinearImpulse(math.random(-100,100), math.random(-60,60))
  else
    patron.body:setSleepingAllowed(true)
    patron.body:setAwake(false)
  end
  patron.shape = love.physics.newCircleShape( math.ceil(extra.scale * 28) ) --the player's shape has a radius of 20
  patron.fixture = love.physics.newFixture(patron.body, patron.shape, 13) -- Attach fixture to body and give it a density of 1.
  patron.fixture:setRestitution(0.1) --let the player bounce
  patron.fixture:setCategory(1)
  
  patron.fixture:setUserData(extra)
  table.insert(objects.patrons, patron)
end

local function enqueuePatrons(dt)
  for i=1, #objects.patrons do
    if not objects.patrons[i] then break end
    local x,y = objects.patrons[i].body:getPosition()
    if math.abs(y - objects.player.body:getY()) < world_height and math.abs(x - objects.player.body:getX()) < world_width then
      objects.patrons[i].body:setAwake(true)

      local vx,vy = objects.patrons[i].body:getLinearVelocity()
      local lineX,lineY = unpack(objects.patrons[i].fixture:getUserData().linePosition)
      local dx = (lineX - x) * 3
      local dy = (lineY - y) * 3
      local mag = math.abs(vx) + math.abs(vy)
      local dMag = math.abs(dx)
      if mag < 15 then
        objects.patrons[i].fixture:getUserData().animation:reset()
      else
        objects.patrons[i].fixture:getUserData().animation:setSpeed(mag/200)
      end
      if dMag > 30 then
        objects.patrons[i].fixture:getUserData().angle = math.atan2(vx, -vy)
      end
      objects.patrons[i].fixture:getUserData().animation:update(dt)
      objects.patrons[i].body:applyForce(dx, dy)
      table.insert(patronsToDraw, objects.patrons[i])
    else
      -- objects.patrons[i].body:setAwake(false)
      if y <= 0 or objects.player.body:getY() + (world_height*2) < y then
        local culled = table.remove(objects.patrons, i)
        culled.body:destroy()
      end
    end
  end

end


local function initializePatrons(first,last)
  for y=first,last do
    if patron_locations[y] then
      for i,data in ipairs(patron_locations[y]) do
        createPatron((data[1] * 32) + math.random(0,10)-5,(y*32) + math.random(0,10)-5, walkCycles[math.random(3)], data[2])
      end
      patron_locations[y] = {}
    end
  end
end

local function makeBounds(width,height)
  local topLeft,topRight,bottomLeft,bottomRight,horizontal,vertical
  topLeft = {0,32}
  topRight = {width*32,0}
  bottomLeft = {0, height*32}
  bottomRight = {width*32, height*32}

  horizontal = {0,0,width*32,0}
  vertical =  {0,0,0,height*32}

  local body = love.physics.newBody(world, unpack(topLeft))
  local shape = love.physics.newEdgeShape(unpack(horizontal))
  local fixture = love.physics.newFixture(body,shape)
  fixture:setCategory(9)

  body = love.physics.newBody(world, unpack(topLeft))
  shape = love.physics.newEdgeShape(unpack(vertical))
  fixture = love.physics.newFixture(body,shape)
  fixture:setCategory(5)

  body = love.physics.newBody(world, unpack(bottomLeft))
  shape = love.physics.newEdgeShape(unpack(horizontal))
  fixture = love.physics.newFixture(body,shape)
  fixture:setCategory(5)

  body = love.physics.newBody(world, unpack(topRight))
  shape = love.physics.newEdgeShape(unpack(vertical))
  fixture = love.physics.newFixture(body,shape)
  fixture:setCategory(5)
end

local function drawFloors()
  love.graphics.setColor(255,255,255)
  love.graphics.drawq(curbL, sidewalkQuad, 0, 0)
  love.graphics.drawq(curbR, sidewalkQuad, (map:getWidth()*32) - 64, 0)
end

local function drawWalls()
  
  
  love.graphics.draw(streetBatch)
  -- drawFloors()
  love.graphics.setColor(255,255,255)
  local leftScale,rightScale,leftX
  leftScale = math.min(math.max(objects.player.body:getX()/600,0.1),1.3)
  rightScale = math.min(math.max((map:getWidth()*32 - objects.player.body:getX())/600,0.1),1.3)
  leftX = -leftScale*64
  rightX = map:getWidth()*32
  love.graphics.draw(bathrooms, -32, -110)
  love.graphics.drawq(streetLeft, buildingsQuad, leftX, -128, 0, leftScale, 4, 0, 0, 0, 0.5)
  love.graphics.drawq(streetRight, buildingsQuad, rightX, 0, 0, rightScale, 4, 0, 0, 0, -0.5)
  love.graphics.drawq(streetRoofLeft, roofQuad, leftX - 256, -128, 0, 1, 4)
  love.graphics.drawq(streetRoofRight, roofQuad, rightX + (64*rightScale), -128, 0, 1, 4)
end


local function draw_hud()
  if gamePaused then
    love.graphics.print("press p to unpause", half_width - 160,half_height)
  end

  local fartsLeft = objects.player.remainingFarts/objects.player.maxFarts
  
  local score_mask = love.image.newImageData('gfx/life_bar_mask.gif')
  local alpha = 0
  
  local function mapPixels(x,y,r,g,b,a)   
    if a > 0 then
      if g <= (fartsLeft*115) + 140 then
        r = 0
        b = 0
        g = 255*fartsLeft
        a = 200
        alpha = g
      else
        r = 0
        a = 0
      end
    end
    return r,g,b,a
  end
  score_mask:mapPixel(mapPixels)
  local maskedScore = love.graphics.newImage(score_mask)
  local elapsed = currentTime - startTime
  local minutes = math.floor(elapsed/60)
  local seconds = elapsed % 60
  local msec = (seconds % 1) * 1000

  -- love.graphics.print(love.timer.getFPS().."fps   "..string.format("% i",math.floor(fartsLeft*100)).."%  "..minutes..string.format("%0i:%02d.%03d", minutes, seconds,msec).." "..disgustCount..", "..collisionCount,5,5)
  love.graphics.setColor(50,255,50)
  love.graphics.print(string.format("%02i:%02d.%03d", minutes, seconds,msec),5,5)
  love.graphics.setColor(255,255,255)
  love.graphics.draw(lower_gi, world_width - 58, 58, 0, 1, 1,48,48)
  love.graphics.draw(maskedScore, world_width - 58, 58, 0, 1, 1,48,48)
  love.graphics.setColor(50,255,50,alpha)
  love.graphics.print('z', world_width - 90, 80)
  love.graphics.setColor(255,255,255)
  if countingDown then
    love.graphics.scale(3)
    if math.ceil(countingDown) > 0 then
      love.graphics.print(math.ceil(countingDown), 126, 16)
    else
      love.graphics.print('go!',112,16)
    end
  end

  if gameOver then
    love.graphics.print("pedestrians disgusted: "..disgustCount, 64, 400)
    love.graphics.print("pedestrians bumped: "..collisionCount, 64, 430)
    love.graphics.print("final time: "..string.format("%02i:%02d.%03d", minutes, seconds,msec), 64, 460)
    love.graphics.print('r to reset', 64, 500)
    love.graphics.print('esc for menu', 64, 540)
    love.graphics.scale(3)
    local sinTime,cosTime = math.abs(math.sin(love.timer.getMicroTime())), math.abs(math.cos(love.timer.getMicroTime()))
    love.graphics.setColor(155*sinTime,255*cosTime,155*sinTime)
    love.graphics.print('you made it!', 20, 20)
    love.graphics.setColor(255,255,255)
  end
end

local function draw_world()
  drawWalls()
  for i=1, #patronsToDraw do
    if patronsToDraw[i].body then
      love.graphics.setColor(patronsToDraw[i].fixture:getUserData().color)
      patronsToDraw[i].fixture:getUserData().animation:draw(patronsToDraw[i].body:getX(), patronsToDraw[i].body:getY(), patronsToDraw[i].fixture:getUserData().angle, 1,1, 32, 32) --objects.patrons[i].scale, 128 * objects.patrons[i].scale, 128 * objects.patrons[i].scale)
    end
  end
  love.graphics.setColor(255,255,255)
  for i=1, #objects.farts do
    if objects.farts[i].body then
      love.graphics.draw(objects.farts[i].particles, objects.farts[i].body:getX(), objects.farts[i].body:getY())
    end
  end

  for i,powerup in pairs(powerups) do
    if powerup then
      love.graphics.draw(powerup.image, powerup.body:getX()- 16, powerup.body:getY() - 16 + math.sin(love.timer.getMicroTime()*15)*5, 0, 2, 2)
    end
  end

  if gameOver then return end
  playerLegs:draw(objects.player.body:getX(), objects.player.body:getY(), objects.player.angle,objects.player.scale,objects.player.scale,32,32)
  if isFarting then
    love.graphics.draw(fartFrame, objects.player.body:getX(), objects.player.body:getY(), objects.player.angle + math.pi,objects.player.scale,objects.player.scale,32,32)
  else
    playerBody:draw(objects.player.body:getX(), objects.player.body:getY(), objects.player.angle,objects.player.scale,objects.player.scale,32,32)
  end
end

function game:init()
  
  love.physics.setMeter(64) --the height of a meter our worlds will be 64px


  world_width = 800
  world_height = 600
  half_width = world_width/2
  half_height = world_height/2
  world = love.physics.newWorld(0,0,world_width,world_height,true)
  objects = {} -- table to hold all our physical objects
  world:setCallbacks(collide)

  -- bloom:send('image_size', {800,600})
  disgustSounds = {}
  for i=1,6 do
    local disgust1 = love.audio.newSource('audio/disgust.wav', 'static')
    table.insert(disgustSounds, disgust1)
  end
  

  fartSound = love.audio.newSource('audio/fart_3.wav', "static")
  --let's create a player
  local playerX,playerY

  local img = love.graphics.newImage('gfx/player_body.gif')
  playerBody = newAnimation(img, 64, 64, 0.1, 6)
  local img = love.graphics.newImage('gfx/player_legs.gif')
  playerLegs = newAnimation(img, 64, 64, 0.1, 6)
  curbL = love.graphics.newImage('gfx/curb_l.gif')
  curbR = love.graphics.newImage('gfx/curb_r.gif')
  street = love.graphics.newImage('gfx/street.gif')
  sidewalk = love.graphics.newImage('gfx/sidewalk.gif')

  curbL:setWrap('repeat', 'repeat')
  curbR:setWrap('repeat', 'repeat')
  street:setWrap('repeat', 'repeat')
  sidewalk:setWrap('repeat', 'repeat')
  
  fartFrame = love.graphics.newImage('gfx/fart_frame.gif')
  fartImages = { love.graphics.newImage('gfx/farticles1.png'), love.graphics.newImage('gfx/farticles2.png'), love.graphics.newImage('gfx/farticles3.png')}
  lower_gi = love.graphics.newImage('gfx/lower_gi.png')
  streetTiles = love.graphics.newImage('gfx/simple_street_tiles.png')
  --let's create a player
  
  powerupImages = {
    hotdog = love.graphics.newImage('gfx/food_hotdog.png'),
    taco = love.graphics.newImage('gfx/food_taco.png'),
    curry = love.graphics.newImage('gfx/food_curry.png'),
    beer = love.graphics.newImage('gfx/food_beer.png'),
    medicine = love.graphics.newImage('gfx/food_gas_no.png')
  }

  map = love.image.newImageData('gfx/map_1.bmp')
  tiles_map = love.image.newImageData('gfx/tiles_map.bmp')

  streetBatch = love.graphics.newSpriteBatch(streetTiles, map:getWidth() * map:getHeight() * 32)
  mapQuads = {}
  for i=0,5 do
    mapQuads[i] = love.graphics.newQuad(i*16, 0, 16, 16, 96, 16)
  end

  walkCycles = {love.graphics.newImage('gfx/walkcycle_1.png'),love.graphics.newImage('gfx/walkcycle_2.png'),love.graphics.newImage('gfx/walkcycle_3.png')}
  -- love.graphics.setCaption(patronCount)
  -- initializePatrons(1,1)
  -- table.foreach(patron_locations, function(i, location)
  --     createPatron(location[1]*28 + math.random(0,10)-5,location[2]*28 + math.random(0,10)-5)
  --   end
  -- )

  streetLeft = love.graphics.newImage('gfx/longstreet.gif')
  streetRight = love.graphics.newImage('gfx/longstreetR.gif')
  streetRoofLeft = love.graphics.newImage('gfx/longstreetRoof.gif')
  streetRoofRight = love.graphics.newImage('gfx/longstreetRRoof.gif')
  bathrooms = love.graphics.newImage('gfx/bathrooms.png')
  streetRight:setWrap('clamp', 'repeat')
  streetLeft:setWrap('clamp', 'repeat')
  streetRoofRight:setWrap('clamp', 'repeat')
  streetRoofLeft:setWrap('clamp', 'repeat')
  buildingsQuad = love.graphics.newQuad(0,0,64,map:getHeight()*32,64,1024)
  roofQuad = love.graphics.newQuad(0,0,256,map:getHeight()*32,3,1024)
  worldBounds = {0,0,map:getWidth()*32,0,map:getWidth()*32,map:getHeight()*32,0,map:getHeight()*32}
  -- sidewalkQuad = love.graphics.newQuad(0,0,64,map:getHeight()*32,64,64)
  -- streetQuad = love.graphics.newQuad(0,0,(map:getWidth()*32) - 256, map:getHeight()*32, 64,64)
  makeBounds(map:getWidth(), map:getHeight())

    --initial graphics setup
  love.graphics.setBackgroundColor(0,0,0) --set the background color to a nice blue
end

function game:enter(state)
  streetBatch:clear()
  startTime = love.timer.getTime()
  currentTime = startTime
  countingDown = 3
  gamePaused = false
  gameOver = false
  lastX = nil
  disgustCount = 0
  collisionCount = 0
  powerups = {}
  activeEffects = {}
  patron_locations = {}
  if objects.patrons then
    for i,patron in pairs(objects.patrons) do
      patron.body:destroy()
    end
  end
  objects.patrons = {}
  local patronCount = 0
  local r,g,b,a
  for x = 0, map:getWidth() - 1 do
    for y = 0, map:getHeight() - 1 do
      r,g,b,a = map:getPixel(x,y) -- Entities
      if r == 255 and b == 0 then
        if patron_locations[y] then
          table.insert(patron_locations[y], {x,g})
        else
          patron_locations[y] = {{x,g}}
        end
        patronCount = patronCount + 1
      elseif b == 255 and r + g == 0 then
        playerX,playerY = x*32,y*32
      elseif b == 255 and r == 255 and g < 255 then
        placePowerup(x*32,y*32, g)
      elseif r + b == 0 then
        placeStaticTile(x*32,y*32, g)
      end

      r,g,b,a = tiles_map:getPixel(x,y)
      if r + g + b == 0 then
        streetBatch:addq(mapQuads[0],x*32,y*32,0,2,2)
      elseif r == 255 and g + b < 100 then
        streetBatch:addq(mapQuads[2],x*32,y*32,0,2,2)
      elseif r == 255 and g + b > 100 then
        streetBatch:addq(mapQuads[3],x*32,y*32,0,2,2)
      elseif r + b > 450 then
        streetBatch:addq(mapQuads[4],x*32,y*32,0,2,2)
      elseif b == 255 then
        streetBatch:addq(mapQuads[5],x*32,y*32,0,2,2)
      else
        streetBatch:addq(mapQuads[1],x*32,y*32,0,2,2)
      end
    end
  end




    -- love.graphics.setCaption(#patron_locations)
  objects.player = {}
  objects.player.activeEffects = {}
  objects.player.angle = 0
  objects.player.speed = 40
  objects.player.scale = 1
  objects.player.fartiness = 0.3
  objects.player.body = love.physics.newBody(world, playerX, playerY, "dynamic") --place the body in the center of the world and make it dynamic, so it can move around
  objects.player.body:setLinearDamping(1)
  objects.player.shape = love.physics.newCircleShape( 7 ) --the player's shape has a radius of 20
  objects.player.fixture = love.physics.newFixture(objects.player.body, objects.player.shape, 2) -- Attach fixture to body and give it a density of 1.
  objects.player.fixture:setRestitution(0.1) --let the player bounce
  objects.player.fixture:setCategory(3)
  objects.player.fixture:setMask(2)
  objects.player.maxFarts = 1
  objects.player.remainingFarts = objects.player.maxFarts

  cam = Camera(objects.player.body:getX(), objects.player.body:getY())
  lowestY = playerY + world_height
  disgustedPatrons = {}
  patronsToDraw = {}

  objects.farts = {}
end

function game:keypressed(key)
  if key == 'p' then
    gamePaused = not gamePaused
  end
  if key == 'r' then
    game:enter()
  end
  if key == 'escape' then
    game:quit()
  elseif key == 'f' then
    love.graphics.toggleFullscreen()
  elseif key == '1' then
    table.insert(objects.player.activeEffects, {type='hotdog', ttl=2})
  elseif key == '2' then
    table.insert(objects.player.activeEffects, {type='beer', ttl=2})
  end
end

function game:keyreleased(key)
  if key == "z" then
    fadeFart = true
  end
end

local function countDown(dt)
  if math.floor(countingDown) == 0 then
    startTime = love.timer.getTime()
    currentTime = startTime
  end
  countingDown = countingDown - dt
end

function game:update(dt)
  if gamePaused then return end
  love.graphics.setCaption(#powerups)
  local fartsLeft = objects.player.remainingFarts/objects.player.maxFarts
  local vx,vy = objects.player.body:getLinearVelocity()
  local zoomLevel = 1
  if math.abs(vy) > 0 then
    zoomLevel = math.min(math.max(250/math.abs(vy),0.8),1)
  end
  cam:zoomTo((zoomLevel+cam.scale)/2)
  cam:lookAt(math.min(math.max(objects.player.body:getX(),300*cam.scale),650*cam.scale), math.min(objects.player.body:getY(),lowestY - half_height))
  local y = math.floor(objects.player.body:getY()/32)
  initializePatrons(math.max(y-20,0),y+20)
  patronsToDraw = {}
  isFarting = false

  table.foreach(disgustedPatrons, function(idx,timer)
    if timer <= 0 then 
      disgustedPatrons[idx] = 0
      return 
    end
    timer = timer - dt
    
    disgustedPatrons[idx] = timer
  end)

  if fadeFart then
    local volume = fartSound:getVolume() - 0.1
    fartSound:setVolume(volume)
    if volume < 0.1 then
      fartSound:stop()
      playingFartSound = false
      fadeFart = false
    end
  end
  world:update(dt) --this puts the world into motion
    if love.keyboard.isDown("z") then
    if not playingFartSound then
      fartSound:setVolume(1.0)
      fartSound:setPitch(fartsLeft)
      fartSound:play()
      playingFartSound = true
    end
    if objects.player.remainingFarts > 0 then
      isFarting = true
      createFart()
    else
      fadeFart = true
    end
    objects.player.remainingFarts = math.max(objects.player.remainingFarts - dt,0)
  else
    objects.player.remainingFarts = math.min(objects.player.remainingFarts + (dt*math.max(0.2,fartsLeft)), objects.player.maxFarts)
  end

  updateFarts(dt)
  enqueuePatrons(dt)
  if countingDown then
    countDown(dt)
    if countingDown > -1 then
      return
    else
      countingDown = false
    end
  end
  if gameOver then return end
  currentTime = currentTime + dt
  --here we are going to create some keyboard events
  playerBody:update(dt)
  playerLegs:update(dt)

  if love.keyboard.isDown("right") then --press the right arrow key to push the player to the right
    objects.player.body:applyForce(objects.player.speed, 0)
  end
  if love.keyboard.isDown("left") then --press the left arrow key to push the player to the left
    objects.player.body:applyForce(-objects.player.speed, 0)
  end
  if love.keyboard.isDown("up") then --press the up arrow key to set the player in the air
    objects.player.body:applyForce(0,-objects.player.speed)
  end
  if love.keyboard.isDown("down") then --press the up arrow key to set the player in the air
    objects.player.body:applyForce(0,objects.player.speed)
  end
  if love.keyboard.isDown('x') then
    objects.player.body:applyForce(-vx,-vy)
  end

  lowestY = math.min(objects.player.body:getY() + world_height,lowestY)
  objects.player.body:setY(math.min(objects.player.body:getY(), lowestY))

  updateWalkSpeed()
  applyEffects(dt)
end

function game:draw()
  
  cam:draw(draw_world)

  draw_hud()
end

function game:focus(f)
  gamePaused = not f or gamePaused
end

return game
